#include "struct.h"

#include "Box2D.h"
#include "wx/config.h"

#include "sysvars.h"

#include <iostream>
#include <sstream>
#include <string>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/shared_ptr.hpp>

#include <cassert>

// initializes simulation
Sim::Sim(SimSettings& _prefs) : mjoint(NULL), numbots(0), numshots(0), numveg(0), cycles(0), lastrepop(0),
			bm(false), down(false), shotpool(sizeof(Shot)), prefs(_prefs) {
	//shots.reserve(512);

	b2AABB b_aabb;
	b_aabb.lowerBound.Set(-10, -10);
	b_aabb.upperBound.Set(static_cast<float>(prefs.width*0.01f)+10, static_cast<float>(prefs.height*0.01f)+10);
	b2Vec2 gravity(0.0f, 0.0f);
	b_world = new b2World(b_aabb, gravity, true);
	b2BodyDef wallsdef;
	b2Body* walls = b_world->GetGroundBody();
	b2PolygonDef shapedef;
	shapedef.restitution = prefs.elast;

	// left and right walls
	b2Vec2 center;
	center.Set(-5.0f, static_cast<float>(prefs.height)*0.005f);
	shapedef.SetAsBox(5.0f, static_cast<float>(prefs.height)*0.005f+10.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	center.Set(static_cast<float>(prefs.width)*0.01f+5.0f, static_cast<float>(prefs.height)*0.005f);
	shapedef.SetAsBox(5.0f, static_cast<float>(prefs.height)*0.005f+10.0f, center, 0.0f);
	walls->CreateShape(&shapedef);
	// top and bottom walls
	center.Set(static_cast<float>(prefs.width)*0.005f, -5.0f);
	shapedef.SetAsBox(static_cast<float>(prefs.width)*0.005f+10.0f, 5.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	center.Set(static_cast<float>(prefs.width)*0.005f, static_cast<float>(prefs.height)*0.01f+5.0f);
	shapedef.SetAsBox(static_cast<float>(prefs.width)*0.005f+10.0f, 5.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	// set contact listener
	b_world->SetContactListener(&listen);

	// seed random number generator
	srand(time(NULL));

	// populate simulation
	for (std::vector<BotTemplate>::iterator p = prefs.templates.begin(); p != prefs.templates.end(); ++p)
		populate(*p);

	// step a bit to avoid overlapping bots and other bad things
	b_world->Step(2.0f, 100, 100);
}

Sim::~Sim() {
	bots.resize(0);
	while (orphan_shots.begin() != orphan_shots.end()) del_shot(orphan_shots.front());
	//for (unsigned int i = numbots; i--; ) del_bot(i);
	//for (unsigned int i = numshots; i--; ) del_shot(i);
	delete b_world;
}

// populates a sim with a bot template
void Sim::populate(const BotTemplate& t) {
	//Code::ptr code(new Code(t.fname));
	const float b_radius = getradius(t.body);
	for (unsigned int i = t.numbots; i--; ) {
		const float x = static_cast<float>(rand())*(prefs.width-2.0f*getradius(t.body))/RAND_MAX+b_radius,
			y = static_cast<float>(rand())*(prefs.height-2.0f*getradius(t.body))/RAND_MAX+b_radius;
		Bot& temp = bots.at(add_bot(x, y, static_cast<float>(rand())*M_PI*2.0f/RAND_MAX, t.body, t.code));
		temp.nrg = t.nrg;
		temp.color[0] = t.color[0];
		temp.color[1] = t.color[1];
		temp.color[2] = t.color[2];
		temp.name = t.name;
		temp.veg = t.veg;
		temp.fixed = t.fixed;
		temp.novis = t.novis;
		temp.noexec = t.noexec;
		if (t.veg) ++numveg;
		temp.ResetMass();
	}
}

// adds a bot to the sim
size_t Sim::add_bot(float x, float y, float ang, float body, Code::ptr code) {
	// create bot
	Bot* bot = new Bot(this, x, y, ang, body, code);
/*	// zero fields
	bot->shell = 0;
	bot->slime = 0;
	bot->poison = 0;
	bot->venom = 0;
	bot->vtimer = 0;
	bot->waste = 0;
	bot->robage = 0;
	memset(bot->data, 0, sizeof(bot->data));
	// create body
	b2BodyDef bdef;
	bdef.position.Set(x*0.01f, y*0.01f);
	bdef.angle = ang;
	bot->data[s_aim] = bot->data[s_setaim] = 1256-rad2db(ang);
	bdef.allowSleep = true;
	bdef.isSleeping = false;
	bdef.isBullet = true;
	bdef.fixedRotation = true;
	bdef.userData = bot;
	bdef.linearDamping = prefs.friction;
	bdef.angularDamping = 100.0f;
	bot->b_body = b_world->CreateBody(&bdef);
	// create shape
	b2CircleDef sdef;
	sdef.friction = prefs.friction;
	sdef.restitution = prefs.elast;
	bot->radius = getradius(body);
	sdef.radius = bot->radius*0.01f;
	sdef.localPosition.Set(0.0f, 0.0f);
	sdef.filter.groupIndex = 4; // bots have +4, shots -4
	bot->b_shape = bot->b_body->CreateShape(&sdef);
	// set variables
	bot->body = body;
	bot->code = code;*/
	bot->id = bots.size();
//	bot->sim = this;
//	bot->radius_changed = false;
	// etc.
//	bot->ResetMass();
	bots.push_back(bot);
	++numbots;
	//dead_bots.resize(numbots, false);
	return bot->id;
}

// marks a bot as dead
void Sim::del_bot(size_t id) {
//	if (id >= bots.size()) return;
//	bots.at(id).dead = true; // the bot is dead now
//	++numdead;
	Bot& bot = bots.at(id);
	//for (std::set<size_t>::iterator p = bot.shots.begin(); p != bot.shots.end(); ++p) {
	//	if (*p < shots.size() && shots.at(*p).parent == bot.id) {
	//		shots.at(*p).parent_exists = false;
	//	} else if (*p >= shots.size()) {
	//		std::cerr << "!! shot #" << *p << " not in shot array !!";
	//	} else if (shots.at(*p).parent != bot.id) {
	//		std::cerr << "!! shot #" << *p << " has different parent !!";
	//	}
	//}
	//b_world->DestroyBody(bot.b_body);
	if (mjoint != NULL && mjoint->GetBody2()->GetUserData() == &bot) mjoint = NULL;
	//if (bot.veg) --numveg;
	//bot.reset();
	//--numbots;
	if (bm && bmid == bot.id) {
		bm = false;
		down = false;
	}
	if (bot.veg) --numveg;
	--numbots;
	//--numdead;
	if (&bot != &bots.back()) {
		// move bot at end of vector to fill in the old, empty, spot
		// bots.back() -> bot
		Bot& bot2 = bots.back();
		const size_t b_id = bot2.id;
		bots.swap(id, b_id);
		//bot2.id = b_id;
		Bot& bot = bots.at(id);
		bot.id = id;
		if (bm && bmid == b_id) {
			bmid = id;
		}
/*		for (std::set<size_t>::iterator p = bot.shots.begin(); p != bot.shots.end(); ++p) {
//			try {
				Shot& shot = shots.at(*p);
				if (shots.at(*p).parent != b_id) {
					std::cerr << "!! " << *p << " has different parent !!";
				} else
				shots.at(*p).parent = id;
//			} catch (...) {
//				std::cerr << "!! " << *p << " not in shot array !!";
//				throw;
//			}
		}*/
	}
	bots.pop_back();
}

bool Sim::cycle() {
//	if ((rand() & 0xFF) == 0) srand(time(NULL)); // reseed

	if (bots.empty() && numshots == 0) {
		return false; // no need to do anything with an empty bot array
	}

	if (numbots != bots.size()) std::cerr << numbots << " != " << bots.size() << std::endl;
	assert(numbots == bots.size());

	for (size_t i = bots.size(); i--;) {
		Bot& temp = bots.at(i);
		if (temp.id != i) std::cerr << "ERROR";
		// execute code
		if (cycles > 0 && !temp.noexec) {
			if (!temp.docode()) {
				del_bot(i);
				continue;
			}
		}
		const b2Vec2 pos = temp.b_body->GetPosition();
		float aim = db2rad((12560-temp.data[s_setaim]+temp.data[s_aimdx]-temp.data[s_aimsx])%1256);

		// reset data
		if (temp.robage < 32000) ++temp.robage;
		if (temp.vtimer > 1) --temp.vtimer;
		++temp.data[s_timer];
		if (temp.data[s_timer] > 32000) temp.data[s_timer] = -32000;
		temp.data[s_nrg] = static_cast<int>(temp.nrg+0.5f);
		temp.data[s_body] = static_cast<int>(temp.body+0.5f);
		temp.data[s_waste] = static_cast<int>(temp.waste+0.5f);
		temp.data[s_shell] = static_cast<int>(temp.shell+0.5f);
		temp.data[s_slime] = static_cast<int>(temp.slime+0.5f);
		temp.data[s_poison] = static_cast<int>(temp.poison+0.5f);
		temp.data[s_venom] = static_cast<int>(temp.venom+0.5f);


		{
			float x = pos.x*100.0f, y = pos.y*100.0f;
			field2bot(x, y);
			temp.data[s_xpos] = static_cast<int>(x+0.5f);
			temp.data[s_ypos] = static_cast<int>(y+0.5f);
		}
		temp.data[s_aim] = temp.data[s_setaim] = 1256-rad2db(aim);
		temp.data[s_vtimer] = static_cast<int>(temp.vtimer+0.5f);
		temp.data[s_robage] = temp.robage;
		temp.data[s_fixed] = temp.data[s_fixpos] = temp.fixed;
		temp.data[s_myeye] = temp.code->myeye;
		temp.data[s_myshoot] = temp.code->myshoot;

		{
			b2Vec2 vel = 100.0f*temp.b_body->GetLinearVelocity();
			const float mag = vec_rotate(vel, -aim);
			temp.data[s_veldn] = -vel.x;
			temp.data[s_velup] = temp.data[s_vel] = vel.x;
			temp.data[s_veldx] = vel.y;
			temp.data[s_velsx] = -vel.y;
			temp.data[s_velscalar] = mag;
		}

		temp.data[s_aimdx] = 0;
		temp.data[s_aimsx] = 0;
		temp.data[s_up] = 0;
		temp.data[s_dn] = 0;
		temp.data[s_dx] = 0;
		temp.data[s_sx] = 0;
		temp.data[s_strpoison] = 0;
		temp.data[s_strvenom] = 0;
		temp.data[s_mkshell] = 0;
		temp.data[s_mkslime] = 0;
		temp.data[s_strbody] = 0;
		temp.data[s_fdbody] = 0;
		temp.data[s_tie] = 0;
		temp.data[s_repro] = 0;
		temp.data[s_mrepro] = 0;
		temp.data[s_sexrepro] = 0;

		for (shotlist::iterator p = temp.shots.begin(); p != temp.shots.end(); ) {
			//if (&temp != temp.b_body->GetUserData()) std::cerr << "Shot array relocated?\n";
			//assert(&temp == temp.b_body->GetUserData());
			if (++p->age > p->range) {
				// so that the iterator remains valid, ish
				del_shot(*(p++));
				continue;
			}
			// shotdecay = 40 as specified in Shots.bas
			//p->nrg *= atan((p->age/p->range-1.0f)*40.0f)/* /atan(-40.0f) */ * -0.6469135775440834;
			// moved to Bot::hit() with a slightly different formula (with very similar results)
			++p;
		}
	}

	for (size_t i = bots.size(); i--;) {
		Bot& temp = bots.at(i);
		// destroy frozen bots
		if (temp.b_body->IsFrozen()) {
			//std::cerr << "Warning: Bot " << i << " escaped the sim entirely, destroying" << std::endl;
			del_bot(i);
			continue;
		}
		temp.vision();
	}


	//dead_bots.assign(numbots, false);
	//bots.resize(numbots -= numdead);
	//assert(numdead == 0);


	/*assert(numshots == shots.size());
	for (unsigned int i = numshots; i--;) {
		Shot& temp = shots.at(i);
	//	if (&temp != temp.b_body->GetUserData()) std::cerr << "Shot array relocated?\n";
		assert(&temp == temp.b_body->GetUserData());
		if (++temp.age > temp.range) {
			del_shot(i);
			continue;
		}
		// shotdecay = 40 as specified in Shots.bas
		temp.nrg *= atan(temp.age/temp.range*40-40)/atan(-40.0f);
	}*/

	for (shotlist::iterator p = orphan_shots.begin(); p != orphan_shots.end(); ) {
		//if (&temp != temp.b_body->GetUserData()) std::cerr << "Shot array relocated?\n";
		//assert(&temp == temp.b_body->GetUserData());
		if (++p->age > p->range) {
			// so that the iterator remains valid, ish
			del_shot(*(p++));
			continue;
		}
		// shotdecay = 40 as specified in Shots.bas
		//p->nrg *= atan(p->age/p->range*40-40)/atan(-40.0f);
		++p;
	}

	// repopulate veggies
	if (lastrepop == 0 && numveg < prefs.vegrepop_threshold) {
		std::vector<BotTemplate> vegspecies;
		for (size_t i = prefs.templates.size(); i--; ) {
			if (prefs.templates.at(i).veg) {
				vegspecies.push_back(prefs.templates.at(i));
			}
		}
		if (vegspecies.size()) for (unsigned long i = prefs.vegrepop_amount; i--; ) {
			BotTemplate& t = vegspecies.at(randint(0, vegspecies.size()-1));
			const float radius = getradius(t.body),
				x = randf(radius, prefs.width-radius), y = randf(radius, prefs.height-radius);
			Bot& temp = bots.at(add_bot(x, y, randf(0, M_PI*2), t.body, t.code));
			temp.nrg = t.nrg;
			temp.color[0] = t.color[0];
			temp.color[1] = t.color[1];
			temp.color[2] = t.color[2];
			temp.name = t.name;
			temp.veg = t.veg;
			temp.fixed = t.fixed;
			temp.novis = t.novis;
			temp.noexec = t.noexec;
			temp.ResetMass();
			++numveg;
		}
		lastrepop = prefs.vegrepop_cooldown;
	}

	// handle mouse joint
	if (mjoint != NULL && (!down || ((Bot *) mjoint->GetBody2()->GetUserData())->id != bmid ||
				((Bot *) mjoint->GetBody2()->GetUserData())->fixed)) {
		b_world->DestroyJoint(mjoint);
		mjoint = NULL;
	}
	if (down && mjoint == NULL && !bots.at(bmid).fixed) {
		b2MouseJointDef mdef;
		mdef.body1 = b_world->GetGroundBody();
		mdef.body2 = bots.at(bmid).b_body;
		mdef.target.Set(((float) bmx)*0.01f, ((float) bmy)*0.01f);
		mdef.maxForce = bots.at(bmid).Mass()*20.0f;
		mdef.dampingRatio = 0.25f;
		mdef.frequencyHz = 200.0f;
		mdef.collideConnected = true;
		mjoint = (b2MouseJoint*) b_world->CreateJoint(&mdef);
	}
	if (down && mjoint != NULL) {
		mjoint->SetTarget(b2Vec2(((float) bmx)*0.01f, ((float) bmy)*0.01f));
	}

	// physics step
	b_world->Step(0.25f, 10, 2);

	for (std::list<Hit>::iterator p = hits.begin(); p != hits.end(); ++p) --(p->life);
	while (!hits.empty() && hits.front().life <= 0) hits.pop_front(); // expire old hits
	++cycles;
	return true;
}

/* *** somewhat related classes *** */

// convenience inline function - gets a string from a wxConfigBase and converts it to float
inline float ConfigFloat(wxConfigBase* cfg, const wxString& name, const wxString &def) {
	double temp;
	if (!cfg->Read(name, def).ToDouble(&temp)) def.ToDouble(&temp);
	return (float) temp;
}

// settings container for a Sim, ctor loads the settings from wxConfig
SimSettings::SimSettings() {
	wxConfigBase* cfg = wxConfigBase::Get();
	width = cfg->Read(_T("/Sim/Width"), 16000);
	height = cfg->Read(_T("/Sim/Height"), 12000);
	maxnonveg = cfg->Read(_T("/Sim/MaxNonVeg"), 200);
	maxveg = cfg->Read(_T("/Sim/MaxVeg"), 100);
	vegrepop_threshold = cfg->Read(_T("/Sim/VegRepopThreshold"), 70);
	vegrepop_amount = cfg->Read(_T("/Sim/VegRepopAmount"), 20);
	vegrepop_cooldown = cfg->Read(_T("/Sim/VegRepopCooldown"), 20);
	vegfeed = cfg->Read(_T("/Sim/VegFeed"), 100);
	vegfeed_type = cfg->Read(_T("/Sim/FeedType"), NRG_PER_CYCLE);

	friction = ConfigFloat(cfg, _T("/Sim/Friction"), _T("0.1"));
	botdensity = ConfigFloat(cfg, _T("/Sim/Density"), _T("1"));
	elast = ConfigFloat(cfg, _T("/Sim/Elast"), _T("0.7"));
	move_eff = ConfigFloat(cfg, _T("/Sim/MoveEff"), _T("0.8"));
	maxvel = cfg->Read(_T("/Sim/MaxVel"), 100);

	storecost = ConfigFloat(cfg, _T("/Sim/StoreCost"), _T("0.04"));
	movcost = ConfigFloat(cfg, _T("/Sim/MovCost"), _T("0.05"));
	agecost = ConfigFloat(cfg, _T("/Sim/AgeCost"), _T("0.0001"));
	shotcost = ConfigFloat(cfg, _T("/Sim/ShotCost"), _T("2"));
	bodyupkeepcost = ConfigFloat(cfg, _T("/Sim/BodyUpkeepCost"), _T("0.00001"));
	poisoncost = ConfigFloat(cfg, _T("/Sim/PoisonCost"), _T("0.01"));
	venomcost = ConfigFloat(cfg, _T("/Sim/VenomCost"), _T("0.01"));
	slimecost = ConfigFloat(cfg, _T("/Sim/SlimeCost"), _T("0.1"));
	shellcost = ConfigFloat(cfg, _T("/Sim/ShellCost"), _T("0.1"));

	std::string istring(cfg->Read(_T("/Sim/Bots"), _T("")).mb_str());
	if (!istring.empty()) {
		std::istringstream istream(istring, std::istringstream::in);
		boost::archive::text_iarchive iarchive(istream);
		iarchive >> templates;
	}
/*	size_t i, p;
	while ((i = serialized.find(')')) != std::string::npos) {
		wxString temp = serialized.substr(1, i-1); // get the bits in between the ()
		serialized.erase(0, i+1); // cut off the beginning
		BotTemplate bot;
		i = temp.find(',');
		bot.name = temp(0, i);
		p = i;
		i = temp.find(',', i+1);
		bot.fname = temp(p+1, i-p-1);
		p = i;
		i = temp.find(',', i+1);
		bot.code = Code::fromStr(temp(p+1, i-p-1));
		p = i;
		i = temp.find(',', i+1);
		bot.nrg = (float) strtod(temp(p+1, i-p-1).ToAscii(), NULL);
		p = i;
		i = temp.find(',', i+1);
		bot.body = (float) strtod(temp(p+1, i-p-1).ToAscii(), NULL);
		p = i;
		i = temp.find(',', i+1);
		bot.color[0] = (float) strtod(temp(p+1, i-p-1).ToAscii(), NULL);
		p = i;
		i = temp.find(',', i+1);
		bot.color[1] = (float) strtod(temp(p+1, i-p-1).ToAscii(), NULL);
		p = i;
		i = temp.find(',', i+1);
		bot.color[2] = (float) strtod(temp(p+1, i-p-1).ToAscii(), NULL);
		p = i;
		i = temp.find(',', i+1);
		bot.numbots = (int) strtol(temp(p+1, i-p-1).ToAscii(), NULL, 10);
		bot.veg = temp[i+1]=='Y';
		bot.fixed = temp[i+2]=='Y';
		bot.noexec = temp[i+3]=='Y';
		bot.novis = temp[i+4]=='Y';
		templates.push_back(bot);
	}*/
}

// save stored settings to wxConfig
void SimSettings::Save() {
	wxConfigBase *cfg = wxConfigBase::Get();

	cfg->Write(_T("/Sim/Width"), (int) width);
	cfg->Write(_T("/Sim/Height"), (int) height);
	cfg->Write(_T("/Sim/MaxNonVeg"), (int) maxnonveg);
	cfg->Write(_T("/Sim/MaxVeg"), (int) maxveg);
	cfg->Write(_T("/Sim/VegRepopThreshold"), (int) vegrepop_threshold);
	cfg->Write(_T("/Sim/VegRepopAmount"), (int) vegrepop_amount);
	cfg->Write(_T("/Sim/VegRepopCooldown"), (int) vegrepop_cooldown);
	cfg->Write(_T("/Sim/VegFeed"), (int) vegfeed);
	cfg->Write(_T("/Sim/FeedType"), (int) vegfeed_type);

	cfg->Write(_T("/Sim/Friction"), wxString::Format(_T("%.2f"), friction));
	cfg->Write(_T("/Sim/Density"), wxString::Format(_T("%.2f"), botdensity));
	cfg->Write(_T("/Sim/Elast"), wxString::Format(_T("%.2f"), elast));
	cfg->Write(_T("/Sim/MoveEff"), wxString::Format(_T("%.2f"), move_eff));
	cfg->Write(_T("/Sim/MaxVel"), (int) maxvel);

	cfg->Write(_T("/Sim/StoreCost"), wxString::Format(_T("%.2f"), storecost));
	cfg->Write(_T("/Sim/MovCost"), wxString::Format(_T("%.2f"), movcost));
	cfg->Write(_T("/Sim/AgeCost"), wxString::Format(_T("%.6f"), agecost));
	cfg->Write(_T("/Sim/ShotCost"), wxString::Format(_T("%.2f"), shotcost));
	cfg->Write(_T("/Sim/BodyUpkeepCost"), wxString::Format(_T("%.6f"), bodyupkeepcost));
	cfg->Write(_T("/Sim/PoisonCost"), wxString::Format(_T("%.2f"), poisoncost));
	cfg->Write(_T("/Sim/VenomCost"), wxString::Format(_T("%.2f"), venomcost));
	cfg->Write(_T("/Sim/SlimeCost"), wxString::Format(_T("%.2f"), slimecost));
	cfg->Write(_T("/Sim/ShellCost"), wxString::Format(_T("%.2f"), shellcost));

/*	wxString serialized;
	for (std::vector<BotTemplate>::iterator p = templates.begin(); p != templates.end(); ++p) {
		serialized += _T('(');
		serialized += p->name;
		serialized += _T(',');
		serialized += p->fname;
		serialized += _T(',');
		serialized += p->code->toStr();
		serialized += wxString::Format(_T(",%f,%f,%f,%f,%f,%d,%c%c%c%c)"), p->nrg,
				p->body, p->color[0], p->color[1], p->color[2], p->numbots,
				p->veg?'Y':'N', p->fixed?'Y':'N', p->noexec?'Y':'N', p->novis?'Y':'N');
	}*/
	std::ostringstream ostream;
	boost::archive::text_oarchive oarchive(ostream);
	oarchive & templates;
	cfg->Write(_T("/Sim/Bots"), wxString(ostream.str().c_str(), wxConvUTF8));
}

// triggers bot-shot collisions
void ShotListener::Add(const b2ContactPoint* point) {
	//SimObject* t1 = (SimObject*) point->shape1->GetBody()->GetUserData(),
	//	* t2 = (SimObject*) point->shape2->GetBody()->GetUserData();
	//if (t1 == NULL || t2 == NULL) return; // probably a wall
	//if (t1->type == BOT && t2->type == SHOT) {
	//	// ensure correct ordering
	//	SimObject* temp = t2;
	//	t2 = t1;
	//	t1 = temp;
	//}
	b2Shape* shape1 = point->shape1, * shape2 = point->shape2;
	if (shape1->GetFilterData().groupIndex == 4 && shape2->GetFilterData().groupIndex == -4) {
		b2Shape* tmp = shape2;
		shape2 = shape1;
		shape1 = tmp;
	}
	//if (t1->type == SHOT && t2->type == BOT) {
	if (shape1->GetFilterData().groupIndex == -4 && shape2->GetFilterData().groupIndex == 4) {
		Shot& stemp = *(Shot*) shape1->GetBody()->GetUserData();
		Bot& temp = *(Bot*) shape2->GetBody()->GetUserData();
		int oldtype = stemp.shoot;
		if (temp.hit(stemp)) {
			Hit t;
			switch (oldtype) {
				case -1:
					t.color[0] = 1.0f;
					t.color[1] = 0.0f;
					t.color[2] = 0.0f;
					break;
				case -2:
					t.color[0] = 1.0f;
					t.color[1] = 1.0f;
					t.color[2] = 1.0f;
					break;
				case -3:
					t.color[0] = 0.0f;
					t.color[1] = 0.0f;
					t.color[2] = 1.0f;
					break;
				case -4:
					t.color[0] = 0.0f;
					t.color[1] = 0.0f;
					t.color[2] = 0.0f;
					break;
				case -5:
					t.color[0] = 1.0f;
					t.color[1] = 1.0f;
					t.color[2] = 0.0f;
					break;
				case -6:
					t.color[0] = 1.0f;
					t.color[1] = 0.0f;
					t.color[2] = 1.0f;
					break;
				default:
					t.color[0] = temp.color[0];
					t.color[1] = temp.color[1];
					t.color[2] = temp.color[2];
					break;
			}
			t.life = 4;
			t.x = point->position.x*100.0f;
			t.y = point->position.y*100.0f;
			current_sim->hits.push_back(t);
		}
	}
}
