#include "struct.h"

#include <boost/serialization/array.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/nvp.hpp>

template<class Archive> inline void Sim::save(Archive & ar, const unsigned int version, wxProgressDialog& pdialog) const {
	ar & BOOST_SERIALIZATION_NVP(numbots)
	   & BOOST_SERIALIZATION_NVP(cycles)
	   & BOOST_SERIALIZATION_NVP(prefs);

	if (!pdialog.Update(0, _T("Saving bot and shot data"))) throw cancelled_exception;

	unsigned int j = 0;

	// serialize bots
	for (size_t i = 0; i < bots.size(); ++i) {
		const Bot& bot = bots.at(i);
		// save angle, position, velocity
		const float angle = bot.b_body->GetAngle();
		const b2Vec2 pos = bot.b_body->GetPosition(),
				vel = bot.b_body->GetLinearVelocity();
		const size_t bot_numshots = bot.shots.size();
		ar & BOOST_SERIALIZATION_NVP(bot)
		   & boost::serialization::make_nvp("bot_x", pos.x)
		   & boost::serialization::make_nvp("bot_y", pos.y)
		   & boost::serialization::make_nvp("bot_velx", vel.x)
		   & boost::serialization::make_nvp("bot_vely", vel.y)
		   & boost::serialization::make_nvp("bot_angle", angle)
		   & BOOST_SERIALIZATION_NVP(bot_numshots);
		const int j = i*200/bots.size();
		if (!pdialog.Update(j)) throw cancelled_exception;
		for (shotlist::const_iterator p = bot.shots.begin(); p != bot.shots.end(); p++) {
			// save shot, position, velocity
			const b2Vec2 pos = p->b_body->GetPosition();
			const b2Vec2 vel = p->b_body->GetLinearVelocity();
			ar & boost::serialization::make_nvp("shot", *p)
			   & boost::serialization::make_nvp("shot_x", pos.x)
			   & boost::serialization::make_nvp("shot_y", pos.y)
			   & boost::serialization::make_nvp("shot_velx", vel.x)
			   & boost::serialization::make_nvp("shot_vely", vel.y);
			if (!pdialog.Update(j)) throw cancelled_exception;
		}
	}
	size_t orphan_shots_size = orphan_shots.size();
	ar & BOOST_SERIALIZATION_NVP(orphan_shots_size);
	for (shotlist::const_iterator p = orphan_shots.begin(); p != orphan_shots.end(); ++p) {
		// save shot, position, velocity
		const b2Vec2 pos = p->b_body->GetPosition();
		const b2Vec2 vel = p->b_body->GetLinearVelocity();
		ar & boost::serialization::make_nvp("shot", *p)
		   & boost::serialization::make_nvp("shot_x", pos.x)
		   & boost::serialization::make_nvp("shot_y", pos.y)
		   & boost::serialization::make_nvp("shot_velx", vel.x)
		   & boost::serialization::make_nvp("shot_vely", vel.y);
		pdialog.Update(++j);
	}
}

template<class Archive> inline void Sim::load(Archive & ar, const unsigned int version, wxProgressDialog& pdialog) {
	ar & BOOST_SERIALIZATION_NVP(numbots)
	   & BOOST_SERIALIZATION_NVP(cycles)
	   & BOOST_SERIALIZATION_NVP(prefs);

	bmx = bmy = bmid = 0;
	bm = down = false;
	mjoint = NULL;


	// initialize world
	b2AABB b_aabb;
	b_aabb.lowerBound.Set(-10.0f, -10.0f);
	b_aabb.upperBound.Set(((float) prefs.width)*0.01f+10.0f, ((float) prefs.height)*0.01f+10.0f);
	b2Vec2 gravity(0.0f, 0.0f);
	b_world = new b2World(b_aabb, gravity, true);

	b2BodyDef wallsdef;
	b2Body* walls = b_world->GetGroundBody();
	b2PolygonDef shapedef;
	shapedef.restitution = prefs.elast;

	// left and right walls
	b2Vec2 center;
	center.Set(-5.0f, static_cast<float>(prefs.height)*0.005f);
	shapedef.SetAsBox(5.0f, static_cast<float>(prefs.height)*0.005f+10.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	center.Set(static_cast<float>(prefs.width)*0.01f+5.0f, static_cast<float>(prefs.height)*0.005f);
	shapedef.SetAsBox(5.0f, static_cast<float>(prefs.height)*0.005f+10.0f, center, 0.0f);
	walls->CreateShape(&shapedef);
	// top and bottom walls
	center.Set(static_cast<float>(prefs.width)*0.005f, -5.0f);
	shapedef.SetAsBox(static_cast<float>(prefs.width)*0.005f+10.0f, 5.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	center.Set(static_cast<float>(prefs.width)*0.005f, static_cast<float>(prefs.height)*0.01f+5.0f);
	shapedef.SetAsBox(static_cast<float>(prefs.width)*0.005f+10.0f, 5.0f, center, 0.0f);
	walls->CreateShape(&shapedef);

	// set contact listener
	b_world->SetContactListener(&listen);

	if (!pdialog.Update(0, _T("Loading bots and shots"))) throw cancelled_exception;

	// load bots
	for (size_t i = 0; i < numbots; ++i) {
		float x, y, vx, vy, angle;
		size_t bot_numshots;
		Bot& bot = bots.new_end();
		ar & BOOST_SERIALIZATION_NVP(bot)
		   & boost::serialization::make_nvp("bot_x", x)
		   & boost::serialization::make_nvp("bot_y", y)
		   & boost::serialization::make_nvp("bot_velx", vx)
		   & boost::serialization::make_nvp("bot_vely", vy)
		   & boost::serialization::make_nvp("bot_angle", angle)
		   & BOOST_SERIALIZATION_NVP(bot_numshots);
		b2BodyDef bdef;
		bdef.position.Set(x, y);
		bdef.angle = angle;
		bdef.allowSleep = true;
		bdef.isSleeping = false;
		bdef.isBullet = true;
		bdef.fixedRotation = true;
		bdef.userData = &bots.at(i);
		bdef.linearDamping = prefs.friction;
		bdef.angularDamping = 100.0f;
		bot.b_body = b_world->CreateBody(&bdef);
		// create shape
		b2CircleDef sdef;
		sdef.friction = prefs.friction;
		sdef.restitution = prefs.elast;
		sdef.radius = bot.radius*0.01f;
		sdef.localPosition.Set(0.0f, 0.0f);
		sdef.filter.groupIndex = 4;
		bot.b_shape = bot.b_body->CreateShape(&sdef);
		bot.ResetMass();
		bot.b_body->SetLinearVelocity(b2Vec2(vx, vy));
		bot.sim = this;
		bot.id = bots.size()-1;
		if (bot.veg) ++numveg;
		while (bot_numshots--) {
			Shot& shot = *new Shot;
			float x, y, vx, vy;
			ar & BOOST_SERIALIZATION_NVP(shot)
			   & boost::serialization::make_nvp("shot_x", x)
			   & boost::serialization::make_nvp("shot_y", y)
			   & boost::serialization::make_nvp("shot_velx", vx)
			   & boost::serialization::make_nvp("shot_vely", vy);
			b2BodyDef bdef;
			bdef.position.Set(x, y);
			bdef.userData = &shot;
			bdef.allowSleep = false;
			bdef.isSleeping = false;
			bdef.isBullet = false;
			bdef.fixedRotation = true;
			bdef.massData.I = 1.0e+15f;
			bdef.massData.center.SetZero();
			bdef.massData.mass = 1.0e+15f;
			shot.b_body = b_world->CreateBody(&bdef);
			b2CircleDef sdef;
			sdef.friction = 0.0f;
			sdef.radius = 0.01f;
			sdef.filter.groupIndex = -4;
			sdef.isSensor = true;
			shot.b_body->CreateShape(&sdef);
			shot.b_body->SetLinearVelocity(b2Vec2(vx, vy));
			shot.parent = &bot;
			shot.sim = this;
			bot.shots.push_back(shot);
		}
		if (!pdialog.Update(i*200/numbots)) throw cancelled_exception;
	}
	size_t orphan_shots_size;
	ar & BOOST_SERIALIZATION_NVP(orphan_shots_size);
	for (size_t i = 0; i < orphan_shots_size; ++i) {
		Shot& shot = *new Shot;
		float x, y, vx, vy;
		ar & BOOST_SERIALIZATION_NVP(shot)
		   & boost::serialization::make_nvp("shot_x", x)
		   & boost::serialization::make_nvp("shot_y", y)
		   & boost::serialization::make_nvp("shot_velx", vx)
		   & boost::serialization::make_nvp("shot_vely", vy);
		b2BodyDef bdef;
		bdef.position.Set(x, y);
		bdef.userData = &shot;
		bdef.allowSleep = false;
		bdef.isSleeping = false;
		bdef.isBullet = false;
		bdef.fixedRotation = true;
		bdef.massData.I = 1.0e+15f;
		bdef.massData.center.SetZero();
		bdef.massData.mass = 1.0e+15f;
		shot.b_body = b_world->CreateBody(&bdef);
		b2CircleDef sdef;
		sdef.friction = 0.0f;
		sdef.radius = 0.01f;
		sdef.filter.groupIndex = -4;
		sdef.isSensor = true;
		shot.b_body->CreateShape(&sdef);
		shot.b_body->SetLinearVelocity(b2Vec2(vx, vy));
		shot.parent = NULL;
		shot.sim = this;
		orphan_shots.push_back(shot);
	}
}
